import FIFO::*;
import GetPut::*;
import StmtFSM::*;
import BRAM::*;
import Vector::*;

typedef Bit#(8) Byte;

interface RC4;
    method Action setKey(Bit#(128) k);
    method Action stop;
    interface Get#(Byte) prng;
endinterface

(* synthesize *)
module mkRC4(RC4);
    BRAM2Port#(Byte, Byte) mem <- mkBRAM2Server(defaultValue);
    Reg#(Vector#(16, Byte)) key <- mkRegU;
    FIFO#(Byte) stream <- mkFIFO;

    Reg#(Bool) loopDone <- mkRegU;
    Reg#(Byte) i <- mkRegU;
    Reg#(Byte) j <- mkRegU;
    Reg#(Byte) si <- mkRegU;
    Reg#(Maybe#(Byte)) fixLateWrite <- mkRegU;

    function makeReq(wr, addr, data) =
        BRAMRequest{write: wr, address: addr, datain: data,
                    responseOnWrite: False};
    function wrReq = makeReq(True);
    function rdReq(addr) = makeReq(False, addr, ?);

    Stmt stmt = seq
        //--------------------------
        // Key Scheduling Algorithm
        //--------------------------

        action i <= 0; loopDone <= False; endaction

        while(!loopDone)
          action
            mem.portA.request.put(wrReq(i  , i  ));
            mem.portB.request.put(wrReq(i+1, i+1));
            if(i == 254) loopDone <= True;
            i <= i + 2;
          endaction

        action i <= 0; j <= 0; loopDone <= False; endaction

        while(!loopDone)
            seq
              mem.portA.request.put(rdReq(i));
              action
                let siNew <- mem.portA.response.get();
                Byte jNew = j + siNew + key[i & 15];
                mem.portA.request.put(rdReq(jNew));
                si <= siNew;
                j  <= jNew;
              endaction
              action
                let sj <- mem.portA.response.get();
                if(i != j)   // only swap if source != destination
                  begin
                    mem.portA.request.put(wrReq(i, sj));
                    mem.portB.request.put(wrReq(j, si));
                  end
                if(i == 255) loopDone <= True;
                i <= i + 1;
              endaction
            endseq

        //------------------------------------
        // Pseudo Random Generation Algorithm
        //------------------------------------
        action i <= 0; j <= 0; loopDone <= False; endaction

        while(!loopDone)
            seq
              action
                mem.portA.request.put(rdReq(i+1));
                i <= i+1;
              endaction
              action
                let siNew <- mem.portA.response.get();
                Byte jNew = j + siNew;
                mem.portA.request.put(rdReq(jNew));
                si <= siNew;
                j  <= jNew;
              endaction
              action
                let sjNew <- mem.portA.response.get();
                Byte addr = si + sjNew;
                if(addr == j)
                    fixLateWrite <= tagged Valid si;
                else if(addr == i)
                    fixLateWrite <= tagged Valid sjNew;
                else
                    fixLateWrite <= tagged Invalid;
                mem.portA.request.put(rdReq(addr));
                mem.portB.request.put(wrReq(i, sjNew));
              endaction
              action
                let prn <- mem.portA.response.get();
                mem.portA.request.put(wrReq(j, si));
                stream.enq(fromMaybe(prn, fixLateWrite));
              endaction
            endseq
        
        stream.clear;

    endseq;

    FSM fsm <- mkFSM(stmt);

    method Action stop;
        loopDone <= True;
    endmethod

    method Action setKey(Bit#(128) k);
        key <= unpack(k);
        fsm.start;
    endmethod

    interface Get prng = fifoToGet(stream);
endmodule


